/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

// @ts-ignore untyped metro
import Metro from 'metro';
// @ts-ignore untyped metro
import {Terminal} from 'metro-core';
// @ts-ignore untyped metro
import { mergeConfig } from 'metro-config'

import path from 'path';
import {
  createDevServerMiddleware,
  indexPageMiddleware,
} from '@react-native-community/cli-server-api';

import enableWatchMode from './watchMode';
import getDefaultConfig, { generateTransformerFile } from '../matrix-config/getDefaultConfig'
import { DEFAULT_HOST } from '../constants'
import HyTerminalReporter from '../lib/HyTerminalReporter'
import { MetroConfig, } from '../matrix-config/configTypes'

export interface MatrixServerConfig {
  projectRoot: string

  metroConfig?: MetroConfig
  onRequest?: (queryUrl: string) => void | Promise<void>
  interactive?: boolean

  onReady?: () => void | Promise<void>
  onError?: () => void | Promise<void>
  customLogReporterPath?: any
  host?: string
  port?: number
  https?: string
  cert?: string
  key?: string
}

export default async function matrixServer(config: MatrixServerConfig) {
  let reportEvent: ((event: any) => void) | undefined;
  const terminal = new Terminal(process.stdout);
  const ReporterImpl = getReporterImpl(config.customLogReporterPath);
  const terminalReporter = new ReporterImpl(terminal, reportEvent);

  const defaultConfig = getDefaultConfig({ projectRoot: config.projectRoot, mode: 'dev' })
  const babelTransformerPath = generateTransformerFile({
    projectRoot: config.projectRoot,
    babelPlugins: config.metroConfig?.transformer?.babelPlugins,
    hyextTransformerConfig: config.metroConfig?.transformer?.hyextTransformerConfig,
  })

  const finalConfig = mergeConfig(
    defaultConfig,
    {
      reporter: terminalReporter,
      ...(config.metroConfig || {}),
      transformer: {
        babelTransformerPath,
      },
    })

  const {middleware, attachToServer} = createDevServerMiddleware({
    host: config.host || DEFAULT_HOST,
    port: finalConfig.server.port,
    watchFolders: finalConfig.watchFolders,
  });
  middleware.use(indexPageMiddleware);

  const customEnhanceMiddleware = finalConfig.server.enhanceMiddleware;
  finalConfig.server.enhanceMiddleware = (
    metroMiddleware: any,
    server: unknown,
  ) => {
    if (customEnhanceMiddleware) {
      metroMiddleware = customEnhanceMiddleware(metroMiddleware, server, config.onRequest);
    }
    return middleware.use(metroMiddleware);
  };

  const serverInstance = await Metro.runServer(finalConfig, {
    host: config.host || DEFAULT_HOST,
    secure: config.https || false,
    secureCert: config.cert,
    secureKey: config.key,
    hmrEnabled: true,
    onReady: config.onReady,
    onError: config.onError,
  });

  const {messageSocket, eventsSocket} = attachToServer(serverInstance);

  terminalReporter.setReportEvent(eventsSocket.reportEvent)

  if (config.interactive !== false) {
    enableWatchMode(messageSocket, config.customLogReporterPath === undefined);
  }

  serverInstance.keepAliveTimeout = 30000;

  return serverInstance
}

function getReporterImpl(customLogReporterPath: string | undefined) {
  if (customLogReporterPath === undefined) {
    return HyTerminalReporter
  }
  try {
    // First we let require resolve it, so we can require packages in node_modules
    // as expected. eg: require('my-package/reporter');
    return require(customLogReporterPath);
  } catch (e) {
    if (e.code !== 'MODULE_NOT_FOUND') {
      throw e;
    }
    // If that doesn't work, then we next try relative to the cwd, eg:
    // require('./reporter');
    return require(path.resolve(customLogReporterPath));
  }
}
